/**
 * \file  GUWebotsDriver.h
 *  GUWebotsDriver
 *  Created by
 *  \author Vlad Estivill-Castro
 *  \date 29/09/2014.
 */

/*
 *
 * This is module acts as a bridge between the whiteboard and a LEGO Mindstorm
 * WEBOTS. Sensors can be commanded to post updates and actuators can be
 * controlled. Currently the module only supports Webots camera and motors.
 *
 *  Copyright 2012 Vlad Estivill-Castro. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *        This product includes software developed by Rene Hexel.
 *
 * 4. Neither the name of the author nor the names of contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * -----------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or
 * modify it under the above terms or under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/
 * or write to the Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#ifndef GUWEBOTSINTERFACE
#define GUWEBOTSINTERFACE

#include <sstream>

#include <webots/Robot.hpp>
#include <webots/DifferentialWheels.hpp>
#include <webots/Camera.hpp>

/**
 * the definition of the status/control messages for differential robot
 */
#include "webots_epuck_nxt_differential_robot/WebotsEPuckNXTdifferentialRobotControlStatus.h"

// we will use the class oriented whiteboard
#include "guwhiteboardwatcher.h"

const float TOP_MOTOR_POWER =100.0;
/**
 *  Webots constants
 */
const int  TIME_STEP= 32; // [ms] // time step of the simulation
const int WHEEL_SAMPLING_PERIOD= 30;        
const int  WEBOTS_PIXELS_MAX_VALUE=255;
const int  TIME_STEP_CAM=64;

/**
 *  Class that acts as a webots controller (a driver).
 *  Webots uses the term controller, we use the term driver.
 *  We will use it mainly for the e-puck
 *  It does inherit from a robot of type  webots::DifferentialWheels
 */
class GUWebotsDriver : public webots::DifferentialWheels {

public:
	/**
	 * GUWebotsDriver constructor 
	 */
    GUWebotsDriver(int8_t theRobotID);

	/**
	 * GUWebotsDriver destructor 
	 */
    virtual ~GUWebotsDriver();

	/**
	 * infinite loop that reads sensor values, report them if necessary and sends commands.
	 * General architecture is to read all sensors with the WEBOTS API
         * Check which ones to report, and build the whiteboard message needed to post
	 */
    void run();

    /**
    * announce we are running
    */
    std::string banner() { return std::string("(c) Vlad Estivill_Castro, webots driver/controller differential robot e-Puck");}

    /*!
     * Better to use a method to analyze control message with a get-Message in out time-triggered approach
     * Will update out status for turning sensor on off, or signals to actuators
     */
    virtual void analyzeControl ();

private :
    bool isRunning;

    /* to keep my status */

    guWhiteboard :: WebotsEPuckNXTdifferentialRobotControlStatus a_robot_status;
    guWhiteboard :: WebotsEPuckNXTdifferentialRobotStatus_t status_Handler;

    /*
     * Always take the parameter as a value in the range [-100,100]
     */
    int16_t bounddParameter(int16_t theParamter) {
         int16_t sign = theParamter >= 0.0 ? 1 : -1;
	 int16_t bounded = fabs( theParamter ) > TOP_MOTOR_POWER*1.0 ?  sign * TOP_MOTOR_POWER : theParamter;
	 return bounded;
    }

    /**
     * Webots parameters
     */
    int speedLeft, speedRight;
    int wheelEncoderPeriod;
     /**
      *      * The speed of the e-Puck Webot    
      */
    //int timesRadiansSpeed;


        /**
	 *       The camera of the e-Puck Webot
         */
        webots::Camera * myCamera;
        // buffers  to read the pixels
        int *imagePerChannel[CHANNEL_SENTINEL];
	const unsigned char *the_image;


    void read_image( )
    {
	    the_image= myCamera->getImage();
	    for (int i = 0; i < a_robot_status.the_camera().camera_width(); i++)
		    {   // imageGetBlue selects yellow on the ground very well
		        // imageGetGreen selects magenta very well
		        // imageGetRed selects blue/cyan very well
		          imagePerChannel[BLUE_CHANNEL][i] = WEBOTS_PIXELS_MAX_VALUE-myCamera->imageGetBlue(the_image, a_robot_status.the_camera().camera_width(), i, 0);
		          imagePerChannel[RED_CHANNEL][i] = WEBOTS_PIXELS_MAX_VALUE-myCamera->imageGetRed(the_image, a_robot_status.the_camera().camera_width(), i, 0);
		          imagePerChannel[GREEN_CHANNEL][i] = WEBOTS_PIXELS_MAX_VALUE-myCamera->imageGetGreen(the_image, a_robot_status.the_camera().camera_width(), i, 0);
		          imagePerChannel[GREY_CHANNEL][i] = WEBOTS_PIXELS_MAX_VALUE-myCamera->imageGetGrey(the_image, a_robot_status.the_camera().camera_width(), i, 0);
                     }
    }

    /**
     * Image analysis routine.  This function returns the amount of visible
     * pixels of the given channel, aims at producing a color detection
     * routine
     */
     int16_t find_total(int tab[], int sizeTab, int threshold){
	     int16_t count=0;
	     for (int i=0; i<sizeTab;i++){
		     if (tab[i]>threshold) count++;
	     }
	     return count;

     }
    /**
     * Image analysis routine.  This function returns the position
     *  This function returns the position of the peak contained in the array given
     * in argument (I have hesitation of what it actually does, Vlad)
     */
     int16_t find_middle(int tab[], int sizeTab){
	   
	     int i,j;
	     int *copy = (int *)malloc(sizeof(int)*sizeTab);
	     int mean=0;
	     int nb_best = sizeTab/10;
	     int *index_bests = (int *)malloc(sizeof(int)*nb_best);
	               
	     // copy the tab, calculate the mean and
	     // test if all the values are identical
	     int identical=1;
	     for (i=0; i<sizeTab;i++){
	       copy[i]=tab[i];
	       mean+=tab[i];
	       if (tab[i]!=tab[0]) identical=0;
	      }

	     // exist with middle point if identical
	      if (identical) {   free(copy);   	free(index_bests); return (int16_t) (sizeTab/2);  	      }

       mean/=sizeTab;
                      	          
        // take the best values of the tab
           for (i=0; i<nb_best; i++){
                           index_bests[i]=-1;
                           int index=-1;
                           int max=0;
                           for (j=0; j<sizeTab; j++){
                                       if (max<copy[j] && copy[j]>mean){
                                             max=copy[j];
                                             index=j;
                                          }
                            }
                          index_bests[i]=index;
                          copy[index]=0;
               }
         free(copy);
	 //
         // calculate the position mean of th best values
         int firstMean=0;
         int count=0;
         for (i=0; i<nb_best; i++){
           if(index_bests[i]!=-1) {
                 firstMean+=index_bests[i];
                       count++;
                 }
         }

	 // count does not change send the middle
         if (count==0) { free(index_bests); return (int16_t) (sizeTab/2); }
                   firstMean/=count;

        // eliminate extreme values
        int secondMean=0;
        count=0;
        for (i=0; i<nb_best; i++){
             if (index_bests[i]<firstMean+sizeTab/10 && index_bests[i]>firstMean-sizeTab/10) {
                 count++;
                  secondMean+=index_bests[i];
               }
         }
        free(index_bests);
        if (count==0) return (int16_t) (sizeTab/2);
	return (int16_t) (secondMean/count);
     }

};

#endif // GUWEBOTSINTERFACE
