/**
 * \file  wb_webots_epuck_NXT_differential_robot.h
 *  wb_webots_epuck_NXT_differential_robot.h
 *  Created by
 *  \author Vlad Estivill-Castro
 *  \date 29/09/2014.
 */

/*
 *
 * This is module acts as a bridge between the whiteboard and a LEGO Mindstorm
 * WEBOTS. Sensors can be commanded to post updates and actuators can be
 * controlled. Currently the module only supports Webots camera and motors.
 *
 *  Copyright 2012 Vlad Estivill-Castro. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *        This product includes software developed by Rene Hexel.
 *
 * 4. Neither the name of the author nor the names of contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * -----------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or
 * modify it under the above terms or under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/
 * or write to the Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#ifndef _WWEBOTSEPUCKDIFFERENTIALNXT_
#define _WWEBOTSEPUCKDIFFERENTIALNXT_

#include <sys/types.h>
#include <gu_util.h>

        /** Enumeration of the Sensor ports available in the NXT differential robot
	 *         */
        enum NXT_Sensor_Ports {
		                NXT_PORT_1 = 0,
		                NXT_PORT_2 = 1,
		                NXT_PORT_3 = 2,
		                NXT_PORT_4 = 3
		        };

        /** Enumeration of the Colour Channels available in camera of the E-Puck differential robot
	 *         */
        enum CAMERA_E_PUCK_CHANNELS {
		                BLUE_CHANNEL = NXT_PORT_1,
		                RED_CHANNEL = NXT_PORT_2,
		                GREEN_CHANNEL = NXT_PORT_3,
		                GREY_CHANNEL = NXT_PORT_4,
                                CHANNEL_SENTINEL=4
		        };

/* The data structure for the values  of a webotsEPuckNXTdifferentialRobot OO-message
 *
 */

struct wb_motor
{
	/********** The motor Power : a percentage from [-100,100], signed, negative is backwards */
	PROPERTY(int16_t, motor_power)
#ifdef __cplusplus
        /** default constructor */
        wb_motor() :  _motor_power(0) {}
        /** copy constructor */
        wb_motor(const wb_motor &other) :  _motor_power(other._motor_power) {}
        /** assignment operator */
	wb_motor &operator=(const  wb_motor &other) {_motor_power=other._motor_power;  return *this;}
#endif
};

struct wb_button
{
	/* true = pressed, false, no contact */
	PROPERTY(bool,pressed)
	/* true = post regularly to the whiteboard, false, no reporting, sensor is off */
	PROPERTY(bool,on)
#ifdef __cplusplus
        /** default constructor */
        wb_button() : _pressed(false), _on(false)  {}
        /** copy constructor */
        wb_button(const wb_button &other) :  _pressed(other._pressed), _on(other._on) {}
        /** assignment operator */
	wb_button &operator=(const  wb_button &other) {_pressed=other._pressed; _on=other._on;  return *this;}
#endif
};

struct wb_distance
{
	/* NXT is mm, Webots ePuck not know TODO */
	PROPERTY(int16_t,distance)
	/* true = post regularly to the whiteboard, false, no reporting, sensor is off */
	PROPERTY(bool,on)
#ifdef __cplusplus
        /** default constructor */
        wb_distance() :  _distance(0), _on(false)  {}
        /** copy constructor */
        wb_distance(const wb_distance &other) :  _distance(other._distance), _on(other._on) {}
        /** assignment operator */
	wb_distance &operator=(const  wb_distance &other) {_distance=other._distance; _on=other._on;  return *this;}
#endif
};

struct wb_encoder
{
	/* Turns of a motor */
	PROPERTY(int16_t,rotations)
	/* true = post regularly to the whiteboard, false, no reporting, sensor is off */
	PROPERTY(bool,on)
#ifdef __cplusplus
        /** default constructor */
        wb_encoder() :  _rotations(0), _on(false) {}
        /** copy constructor */
        wb_encoder(const wb_encoder &other) :  _rotations(other._rotations), _on(other._on) {}
        /** assignment operator */
	wb_encoder &operator=(const  wb_encoder &other) {_rotations=other._rotations; _on=other._on;  return *this;}

#endif
};

struct wb_camera
{
	/* camera width */
	PROPERTY(u_int8_t,camera_width)
	/* camera height */
	PROPERTY(u_int8_t,camera_height)
	/* camera camera_threshold */
	PROPERTY(u_int8_t,camera_threshold)
	/* camera median per channel */
	ARRAY_PROPERTY(u_int8_t, median, (GREY_CHANNEL+1))
	/* camera total per channel */
	ARRAY_PROPERTY(u_int8_t, totalpixel_count, (GREY_CHANNEL+1))
	/* true = post regularly to the whiteboard, false, no reporting, sensor is off */
	PROPERTY(bool,on)
#ifdef __cplusplus
        /** default constructor */
        wb_camera() :  _camera_width(0), _camera_height(0), _camera_threshold(0), _on(false) {}
        /** copy constructor */
        wb_camera(const wb_camera &other) :  	_camera_width(other._camera_width),
       						_camera_height(other._camera_height), 
						_camera_threshold(other._camera_threshold),
					      	_on(other._on) {
							for (CAMERA_E_PUCK_CHANNELS channel_i= BLUE_CHANNEL; 
									channel_i != GREY_CHANNEL ;
									channel_i = static_cast<CAMERA_E_PUCK_CHANNELS>(static_cast<int>(channel_i) + 1))
									{ _median[channel_i]=other._median[channel_i];
									  _totalpixel_count[channel_i]=other._totalpixel_count[channel_i];
									}
						}
        /** assignment operator */
	wb_camera &operator=(const  wb_camera &other) { _camera_width=other._camera_width;
       							_camera_height=other._camera_height; 
							_camera_threshold=other._camera_threshold;
							for (CAMERA_E_PUCK_CHANNELS channel_i= BLUE_CHANNEL; 
									channel_i != GREY_CHANNEL ;
									channel_i = static_cast<CAMERA_E_PUCK_CHANNELS>(static_cast<int>(channel_i) + 1))
									{ _median[channel_i]=other._median[channel_i];
									  _totalpixel_count[channel_i]=other._totalpixel_count[channel_i];
									}
	                                        	_on=other._on;  return *this;}
#endif
};


struct wb_webots_epuck_NXT_differential_robot
{
        PROPERTY(int16_t ,robot_ID)
	/* Differential robot has two motors with encoders */
        PROPERTY(struct wb_motor, left_motor)
	PROPERTY(struct wb_motor, right_motor)
        PROPERTY(struct wb_encoder, left_encoder)
	PROPERTY(struct wb_encoder, right_encoder)
	/* for the ePuck */
        PROPERTY(int16_t ,max_times_radians_speed)
	/* Also left and right touch sensors */
        PROPERTY(struct wb_button, left_button)
	PROPERTY(struct wb_button, right_button)
	    /* For now one sonar sensor*/
	PROPERTY(struct wb_distance, sonar_sensor)
	    /* For now one camera*/
	PROPERTY(struct wb_camera, the_camera)
#ifdef __cplusplus
        /** copy constructor */
	wb_webots_epuck_NXT_differential_robot() :
		_robot_ID(),
	       	_left_motor(), _right_motor(),  _left_encoder(), _right_encoder(), _max_times_radians_speed(),  _left_button(), _right_button(), _sonar_sensor(), _the_camera() {}
        /** copy constructor */
        wb_webots_epuck_NXT_differential_robot(const wb_webots_epuck_NXT_differential_robot &other) :  
		_robot_ID(other._robot_ID),
		_left_motor(other._left_motor), _right_motor(other._right_motor),
		_left_encoder(other._left_encoder), _right_encoder(other._right_encoder),
		_max_times_radians_speed(other._max_times_radians_speed),
		_left_button(other._left_button), _right_button(other._right_button),
		_sonar_sensor(other._sonar_sensor),
		_the_camera(other._the_camera)
       	{}
        /** assignment operator */
	wb_webots_epuck_NXT_differential_robot &operator=(const  wb_webots_epuck_NXT_differential_robot &other) {
		_robot_ID=other._robot_ID;
		_left_motor=other._left_motor; _right_motor=other._right_motor; 
		_left_encoder=other._left_encoder; _right_encoder=other._right_encoder; 
		_max_times_radians_speed=other._max_times_radians_speed;
		_left_button=other._left_button; _right_button=other._right_button; 
		_sonar_sensor=other._sonar_sensor;
		_the_camera=other._the_camera;
	       	return *this;}
#endif
};


#endif //_WWEBOTSEPUCKDIFFERENTIALNXT_
