/**
 * \file  GUWebotsDriver.cpp
 *  GUWebotsDriver
 *  Created by
 *  \author Vlad Estivill-Castro
 *  \date 29/09/2014.
 */

/*
 *
 * This is module acts as a bridge between the whiteboard and a LEGO Mindstorm
 * WEBOTS. Sensors can be commanded to post updates and actuators can be
 * controlled. Currently the module only supports linux over USB connections.
 *
 *
 *  Copyright 2012 Vlad Estivill-Castro. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *        This product includes software developed by Rene Hexel.
 *
 * 4. Neither the name of the author nor the names of contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * -----------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or
 * modify it under the above terms or under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/
 * or write to the Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include "gu_util.h"  // for DBG = DEBUG
#include <iostream>
#include <stdlib.h>

#include "webots_epuck_nxt_differential_robot/GUWebotsDriver.h"

// All the webots classes are defined in the "webots" namespace
using namespace webots;

using namespace std;
using namespace guWhiteboard;


// GUWebotsDriver constructor
GUWebotsDriver:: GUWebotsDriver(int8_t aRobotID): DifferentialWheels()  {

    a_robot_status.set_robot_ID( aRobotID );
    cerr << banner() <<  endl;
    cerr << "ROBOT ID: " << a_robot_status.robot_ID() << endl;

    isRunning=false;

    /********** start with no speed on the motor */
    speedLeft=0; speedRight=0;
    wb_motor my_left_motor; my_left_motor.set_motor_power(static_cast<int16_t>(0));
    wb_motor my_right_motor; my_right_motor.set_motor_power(static_cast<int16_t>(0));
    a_robot_status.set_left_motor(my_left_motor);
    a_robot_status.set_right_motor(my_right_motor);
    a_robot_status.set_right_motor(my_right_motor);
    int16_t theMaxSpeed=static_cast<int16_t>(        round (getMaxSpeed()/3.14159265) ) ;
    a_robot_status.set_max_times_radians_speed ( theMaxSpeed   );

    /********** Set up the camera */
     myCamera= getCamera ("camera");
     myCamera->enable(TIME_STEP_CAM);
    a_robot_status.the_camera().set_camera_width(myCamera->getWidth());
    a_robot_status.the_camera().set_camera_height(myCamera->getHeight());
    a_robot_status.the_camera().set_camera_threshold(0);
   for (CAMERA_E_PUCK_CHANNELS channel_i= CAMERA_E_PUCK_CHANNELS::BLUE_CHANNEL;
                         channel_i != CAMERA_E_PUCK_CHANNELS::CHANNEL_SENTINEL ;
                         channel_i = static_cast<CAMERA_E_PUCK_CHANNELS>(static_cast<int>(channel_i) + 1))
                         { 
				 imagePerChannel[channel_i]= (int *)malloc(sizeof(int)*a_robot_status.the_camera().camera_width());
			}
    a_robot_status.the_camera().set_on(false);

    /********** start with motor in zero and off */
    wheelEncoderPeriod = WHEEL_SAMPLING_PERIOD;
    enableEncoders(wheelEncoderPeriod);
    setEncoders(0.0,0.0);
    wb_encoder my_left_encoder; my_left_encoder.set_rotations(static_cast<int16_t>(0)); my_left_encoder.set_on(false);
    wb_encoder my_right_encoder; my_right_encoder.set_rotations(static_cast<int16_t>(0)); my_right_encoder.set_on(false);
    a_robot_status.set_left_encoder(my_left_encoder);
    a_robot_status.set_right_encoder(my_right_encoder);

    /************* clean the whiteboard with first status message ************/
   status_Handler.set(a_robot_status);
    /************* clean the whiteboard with a command equal to the first status message ************/
    WebotsEPuckNXTdifferentialRobotControl_t        control_Handler;
    WebotsEPuckNXTdifferentialRobotControlStatus    theCommand; 
    theCommand=a_robot_status;
    control_Handler.set(theCommand);

    cerr <<  " CONSTRUCTOR FINISHED :" << endl;
}

// GUWebotsDriver destructor
GUWebotsDriver :: ~GUWebotsDriver() {
    // Enter here exit cleanup code
    //
   for (CAMERA_E_PUCK_CHANNELS channel_i= CAMERA_E_PUCK_CHANNELS::BLUE_CHANNEL;
                         channel_i != CAMERA_E_PUCK_CHANNELS::CHANNEL_SENTINEL ;
                         channel_i = static_cast<CAMERA_E_PUCK_CHANNELS>(static_cast<int>(channel_i) + 1))
                         { free ( imagePerChannel[channel_i] ); 
			 }
}

// User defined function for initializing and running
// the GUWebotsDriver class
void GUWebotsDriver ::run() {

    // Main loop
    int result;
    do {
        //Better to eliminate subscribe and facilitate true idempotent messages
	//Check if there is any control
 	analyzeControl ();	
        // Code to handle the rotation sensors (encoders):
	//
	// Test if encoder is ON
	if ( a_robot_status.left_encoder().on() )
          a_robot_status.left_encoder().set_rotations( static_cast<int16_t>(getLeftEncoder()));

	if ( a_robot_status.right_encoder().on() )
          a_robot_status.right_encoder().set_rotations( static_cast<int16_t>(getRightEncoder()) );
        // READ the Camera
	read_image();
	if (a_robot_status.the_camera().on())
	{ for (CAMERA_E_PUCK_CHANNELS channel_i= CAMERA_E_PUCK_CHANNELS::BLUE_CHANNEL;
                     channel_i != CAMERA_E_PUCK_CHANNELS::CHANNEL_SENTINEL ;
                     channel_i = static_cast<CAMERA_E_PUCK_CHANNELS>(static_cast<int>(channel_i) + 1))
		 { a_robot_status.the_camera().set_totalpixel_count( 
		    find_total( imagePerChannel[channel_i],a_robot_status.the_camera().camera_width(),a_robot_status.the_camera().camera_threshold()),
		    channel_i 
		    );
                   a_robot_status.the_camera().set_median( 
		    find_middle( imagePerChannel[channel_i],a_robot_status.the_camera().camera_width() ),
		     channel_i
		    );
		 }
                cerr <<  " Camera ON :" << endl;
	}

	// Post new status
         status_Handler.set(a_robot_status);

	//Set the speed of the motors
        setSpeed(speedLeft,speedRight);

        // Perform a simulation step of TIME_STEP milliseconds and leave the loop when the simulation is over
        result = step(TIME_STEP);
        usleep(10); //give wb message   a chance to enable/disable devices

    } while (result != -1);
}

void
GUWebotsDriver:: analyzeControl ()
{
	// Lets do a get-Message on our Control and update out internal status accordingly

    WebotsEPuckNXTdifferentialRobotControl_t        control_Handler;
    WebotsEPuckNXTdifferentialRobotControlStatus    theCommand; 
    theCommand=control_Handler.get();

    DBG(cout << " We received a control message " << theCommand.description() <<  std::endl;)

    // we get motor as a power in [0,100] with a sign for direction
    speedLeft= (         bounddParameter( theCommand.left_motor().motor_power()   ) /100.0 ) *(getMaxSpeed()/getSpeedUnit()) ;
    speedRight= (        bounddParameter(theCommand.right_motor().motor_power()   ) /100.0)*(getMaxSpeed()/getSpeedUnit()) ;
  // Turn encoder on and off
    a_robot_status.left_encoder().set_on(   theCommand.left_encoder().on()  );
    a_robot_status.right_encoder().set_on(  theCommand.right_encoder().on() );
  // Turn the camera on and off
    a_robot_status.the_camera().set_on(  theCommand.the_camera().on() );
}
